#!/usr/bin/perl -w

use strict;
use IO::File;
use POSIX;
use Fcntl ':flock';
use Getopt::Long;
use Pod::Usage;

my $sshkey = '/var/lib/passwd.cued/getpwfiles2_rsa';
my $lastupdatefile = '/var/lib/passwd.cued/lastupdate';
my $lockfile = '/var/lib/passwd.cued/getpwfiles.lockfile';

my $proto = 'passwd.proto';
my $gproto = 'group.proto';

my $debug = 0;
my $verbose = 0;
my $force = 0;
my $xusers = 0;
my $protodir = '/etc';
GetOptions (
	'debug'		=> \$debug,
	'verbose'	=> \$verbose,
	'force'		=> \$force,
	'xusers'	=> \$xusers,
	'proto:s'	=> \$protodir,
) or pod2usage ( -verbose => 1) && exit;

my $gfile = $protodir.'/group.cued';
my $pfile = $protodir.'/passwd.cued';
my $sfile = $protodir.'/shadow.cued';
my $cfile = $protodir.'/crsmap';

$SIG{'ALRM'} = \&timeout;

my $alarmdelay = 300;
alarm ($alarmdelay);

umask (022);

print "Debug mode\n" if $debug;
print "Verbose mode\n" if $verbose;
print "Force update\n" if $force;
print "Exclude x-users\n" if $xusers && ($debug || $verbose);
print "Proto directory $protodir\n" unless $protodir eq '/etc';

open (LF, ">$lockfile") or die "Unable to write $lockfile: $!\n";
flock (LF, (LOCK_EX | LOCK_NB)) or die "Unable to flock $lockfile: $!\n";
print "Locked\n" if $debug;

my $lastupdate = 0;
if ( -f $lastupdatefile) {
	open (LU, $lastupdatefile) 
		or die "Unable to read $lastupdatefile: $!\n";
	$lastupdate = <LU> || 0;
	chomp $lastupdate;
	print "Lastupdate $lastupdate\n" if $debug;
	close (LU);
}

my $RAplines = parseproto ($proto);
foreach (@$RAplines) {
	print "$_\n" if $debug;
}
# only generate groups if there is a group.proto file
my $RAgplines;
if ( -f "$protodir/$gproto" ) {
	$RAgplines = parseproto ($gproto);
	foreach (@$RAgplines) {
		print "$_\n" if $debug;
	}
}

my (@pwdata, @gdata, @cdata);
my $mapfile;
my %mapdata;
my $lastchange=0;
my $eod;
my $receiving;
my $target;

$ENV{LANG}='C';
open (SSH, "/usr/bin/ssh -n -T -oStrictHostKeyChecking=no -i $sshkey -q passwdd\@pw-serv|")
	or die "Cannot open ssh to pw-serv-vm: $!\n";
while (<SSH>) {
	chomp;
	die("Received junk after End of Data - aborting\n") if $eod;
	if (/^---/)
	{
		$receiving=1;
		undef($target);
		if (/^---lastchange (\d+)$/) {
			$lastchange = $1;
			print "Lastchange = $lastchange\n" if $debug;
			next;
		}
		if (/^---mapfile:(\S+)---$/) {
			$mapdata{$1}=[];
			$target=$mapdata{$1};
			next;
		}
		$target=\@pwdata, next if /^---passwd---$/;
		$target=\@gdata, next if /^---group---$/;
		$target=\@cdata, next if /^---crsmap---$/;
		$eod=1, last if /^---eod---$/;
	}
	elsif (!$receiving)
	{
		print "$_\n"; # Server can print debug info lines before sending data
		die("Server failed to get lock\n") if /^Failed to obtain lock/;
	}
	die("Received data without target\n$_\n") if !$target;
	push(@$target, $_) if $target;
}
close (SSH);
die("No End of Data marker read - aborting\n") unless $eod;

unless ($force || ($lastchange && $lastchange > $lastupdate)) {
	print "\nCurrent files are up to date\n" if $debug || $verbose;
	exit 0 unless $debug;
}

# Generate crsid map for our own purposes
my %crsmap=map {next unless /^(\S+) (\S+)$/; $1 ne $2 ? ($1,$2) : ()} @cdata;
print map {"$_ -> $crsmap{$_}\n"} sort keys(%crsmap) if $debug;

open (PWOUT, ">$pfile.new") or die "Unable to write $pfile.new: $!\n";
# Create new shadow file with correct permissions from the start
sysopen(SOUT, "$sfile.new", O_WRONLY|O_TRUNC|O_CREAT, 0000)
	or die "Unable to write $sfile.new: $!\n";

# Build a hash based on pwdata of (real) username to various other fields.
my %pwhash;
foreach my $pwline (@pwdata) {
	my ($id, $pw, $uid, $gid, $gcos, $hdir, $shell) =
		split (":", $pwline);
	my $ex = '';
	if ($id =~ /^x-(.*)/) {
		$ex = 'x-';
		$id = $1;
		print "Ex user $id\n" if $debug;
	}
	print "Parsing $id from $pwline\n" if $debug;
	print "$id, $gcos\n" if $debug;
	$pwhash{$id} = { 
			pw => $pw,
			uid => $uid,
			gid => $gid,
			gcos => $gcos,
			hdir => $hdir,
			shell => $shell,
			ex => $ex,
			};
}

# Scan for specific passwd.proto lines first, print specific users and mark
# them as being output.  Lines starting with +user need interpolation.
# Look for first default line and record that too in pwhash for later use
# Lines without a + should be output without reference to pwhash
print "Outputting password lines\n" if $debug;
my $nplines = 0;
foreach my $pline (@$RAplines) {
	print $pline, "\n" if $debug;
	my ($id, $pw, $uid, $gid, $gcos, $hdir, $shell) =
		split (":", $pline);
	my $ex = '';
	die "Bad pline $pline" unless defined $shell;
	if ($id =~ /^\+(.+)$/) {
		$id = $1;
		my $crsid = $crsmap{$id};
		print "pline: Mapping $id to $crsid\n" if $crsid && $debug;
		$id = $crsid if $crsid;
		print "Fill in for $id\n" if $debug;
		unless (defined $pwhash{$id}) {
			warn "Unable to complete $id\n" 
				if ($debug || $verbose);
			next;
		}
		# de-dup in case both engid and crsid lines asked for
		# by proto file (which both map to crsid now)
		next if $pwhash{$id}{'output'};
		# If an ex user:
		# Set $ex string 
		# explicitly blat PW
		# explicitly blat shell
		if ($pwhash{$id}{'ex'}) {
			$ex = $pwhash{$id}{'ex'};
			$pw = '*LK*';
			$shell = '/bin/false';
		}
		# Fill in fields which aren't already specified by proto
		$pw = $pwhash{$id}{'pw'} unless $pw;
		$uid = $pwhash{$id}{'uid'} unless $uid;
		$gid = $pwhash{$id}{'gid'} unless $gid;
		$gcos = $pwhash{$id}{'gcos'} unless $gcos;
		$hdir = $pwhash{$id}{'hdir'} unless $hdir;
		if ($hdir =~ /\$/) {
			my $name = $id;# name is a reasonable alias for id
			my $user = $id;# user is a reasonable alias for id
			print "pline: Interpolating $hdir\n" if $debug;
			$hdir = eval "\"$hdir\"" || $hdir; # revert if false
			print "pline: result $hdir\n" if $debug;
		}
		$shell = $pwhash{$id}{'shell'} unless $shell;
		$pwhash{$id}{'output'} = 1;
		print PWOUT "$ex$id:x:$uid:$gid:$gcos:$hdir:$shell\n";
		print SOUT "$ex$id:${pw}:::::::\n";
		$nplines++;
		next;
	}
	# Deal with default rules
	if ($id eq '+') {
		if (defined $pwhash{'default'}) {
			warn "Default pattern already defined\n";
			next;
		}
		$pwhash{'default'}{'seen'} = 1;	# make sure the hash is defined
		$pwhash{'default'}{'pw'} = $pw if $pw;
		$pwhash{'default'}{'uid'} = $uid if $uid;
		$pwhash{'default'}{'gid'} = $gid if $gid;
		$pwhash{'default'}{'gcos'} = $gcos if $gcos;
		$pwhash{'default'}{'hdir'} = $hdir if $hdir;
		$pwhash{'default'}{'shell'} = $shell if $shell;
		next;
	}
	# Non-interpolated proto lines:
	# Make sure fields are always filled.
	$pw = 'x' unless $pw;
	$uid = 0 unless $uid;
	$gid = 0 unless $gid;
	# Print result
	print PWOUT "$id:x:$uid:$gid:$gcos:$hdir:$shell\n";
	print SOUT "$id:${pw}:::::::\n";
	$pwhash{$id}{'output'} = 1;
	$nplines++;
}

# Now if a default is defined, print remaining users
if (defined $pwhash{'default'}) {
	print "Printing users matching default\n" if $debug;
	# Rescan pwdata, so we preserve order of users from it
	foreach my $pwline (@pwdata) {
		my $id = (split (":", $pwline))[0];
		my $crsid = $crsmap{$id};
		print "default: Mapping $id to $crsid\n" if $crsid && $debug;
		$id = $crsid if $crsid;

		my $ex = '';
		if ($id =~ /^x-(.*)/) {
			$ex = 'x-';
			$id = $1;
			print "Ex user $id\n" if $debug;
			print "Skipping ex user $id\n" if $xusers && $debug;
			next if $xusers;
		}
		next if $pwhash{$id}{'output'};
		print "Default output $id\n" if $debug;
		my ($pw, $uid, $gid, $gcos, $hdir, $shell);
		$pw = $pwhash{'default'}{'pw'} || $pwhash{$id}{'pw'};
		$uid = $pwhash{'default'}{'uid'} || $pwhash{$id}{'uid'};
		$gid = $pwhash{'default'}{'gid'} || $pwhash{$id}{'gid'};
		$gcos = $pwhash{'default'}{'gcos'} || $pwhash{$id}{'gcos'};
		$hdir = $pwhash{'default'}{'hdir'} || $pwhash{$id}{'hdir'};
		$shell = $pwhash{'default'}{'shell'} || $pwhash{$id}{'shell'};
		$ex = $pwhash{$id}{'ex'} || '';
		# Make sure some fields are always filled.
		$pw = 'x' unless $pw;
		$uid = 0 unless $uid;
		$gid = 0 unless $gid;
		$gcos = '' unless $gcos;
		$hdir = '' unless $hdir;
		$shell = '' unless $shell;
		# Do homedir substitutions where appropriate
		if ($hdir =~ /\$/) {
			my $crsid = $crsmap{$id};
			# user and name are reasonable aliases for id
			my $user = $crsid || $id;
			my $name = $user;
			print "Default: Interpolating $hdir\n" if $debug;
			$hdir = eval "\"$hdir\"" || $hdir; # revert if false
			print "Default: result $hdir\n" if $debug;
		}
		print PWOUT "$ex$id:x:$uid:$gid:$gcos:$hdir:$shell\n";
		print SOUT "$ex$id:${pw}:::::::\n";
		$pwhash{$id}{'output'} = 1;
		$nplines++;
	}
}
close (PWOUT) or die "unable to close $pfile.new: $!\n";
close (SOUT) or die "unable to close $sfile.new: $!\n";

my $plines = 0;
if (-f $pfile) {
	open (PW, $pfile) or die "Cannot read $pfile: $!\n";
	while (<PW>) {
		$plines++;
	}
	close (PW);
	if ($nplines + 100 < $plines) {
		warn "New password file more than 100 lines shorter\n";
		die "Exiting\n" unless $force;
	}
}
unless ($debug) {
	print "Installing $pfile\n" if $verbose;
	chmod (0644, "$pfile.new");
	rename ("$pfile.new", $pfile) 
		or die "Rename $pfile.new to $pfile failed: $!\n";
	print "Installing $sfile\n" if $verbose;
	chmod (0000, "$sfile.new");
	rename ("$sfile.new", $sfile) 
		or die "Rename $sfile.new to $sfile failed: $!\n";
}

open (GROUT, ">$gfile.new") or die "Unable to write $gfile.new: $!\n";
my %ghash;
foreach my $gline (@gdata) {
	my ($group, $star, $gid, $members) = split (":", $gline);
	$ghash{$group} = {
		gid => $gid,
		members => $members,
		};
}

my $nglines = 0;
# Same for group.proto file as for passwd.proto file
foreach my $gpline (@$RAgplines) {
	print $gpline, "\n" if $debug;
	my ($group, $star, $gid, $members) = split (":", $gpline);
	if ($group =~ /^\+(.+)$/) {
		$group = $1;
		unless (defined $ghash{$group}) {
			warn "Unable to complete $group\n";
		}
		$gid = $ghash{$group}{'gid'} unless $gid;
		$members = $ghash{$group}{'members'} unless $members;
		print GROUT "$group:*:$gid:$members\n";
		$ghash{$group}{'output'} = 1;
		$nglines++;
		next;
	}
	if ($group eq '+') {
		print "Processing default line $gpline\n" if $debug;
		if (defined $ghash{'default'}) {
			warn "Default group pattern already defined\n";
			next;
		}
		$ghash{'default'}{'seen'} = 1;	# make sure the hash is defined
		$ghash{'default'}{'gid'} = $gid if $gid;
		$ghash{'default'}{'members'} = $members if $members;
		next;
	}
	# Non-interpolated proto lines:
	# Make sure some fields always filled
	$gid = 0 unless $gid;
	$members = '' unless $members;
	print GROUT "$group:*:$gid:$members\n";
	$nglines++;
}

if (defined $ghash{'default'}) {
	print "Printing groups matching default\n" if $debug;
	my ($gid, $members);
	foreach my $group (sort keys %ghash) {
		next if $group eq 'default';
		next if $ghash{$group}{'output'};
		$gid = $ghash{'default'}{'gid'} || $ghash{$group}{'gid'}
				|| 0;
		$members = $ghash{'default'}{'members'} 
				|| $ghash{$group}{'members'} || '';
		if ($members eq '') {
			print GROUT "$group:*:$gid:$members\n";
			$nglines++;
			next;
		} 
		my $count = 0;
		my $gline = "$group:*:$gid:";
		foreach my $member (split ",", $members) {
			$gline .= "$member,";
			$count++;
			if ($count > 9) {
				chop $gline;
				print GROUT "$gline\n";
				$nglines++;
				$gline = "$group:*:$gid:";
				$count = 0;
			}
		}
		chop $gline;
		print GROUT "$gline\n";
		$nglines++;
	}
}
close (GROUT) or die "Unable to close $gfile.new: $!\n";

my $glines = 0;
if ( -f $gfile ) {
	open (GRP, $gfile) or die "unable to read $gfile: $!\n";
	while (<GRP>) {
		$glines++;
	}
	close (GRP);
	if ($nglines + 100 < $glines) {
		warn "New group file more than 100 lines shorter\n";
		die "Exiting\n" unless $force;
	}
}
unless ($debug) {
	print "Installing $gfile\n" if $verbose;
	chmod (0644, "$gfile.new");
	rename ("$gfile.new", $gfile)
		or die "Rename $gfile.new to $gfile failed: $!\n";
}

foreach my $map (keys %mapdata) {
	my $mapfile = "/etc/$map";
	my $nmapfile = "$mapfile.new";
	print "Creating $mapfile.new\n" if $debug;
	open (NMAP, ">$nmapfile") or die "Unable to write $nmapfile: $!\n";
	print NMAP join "\n", @{$mapdata{$map}};
	print NMAP "\n";
	close (NMAP) or die "unable to close $mapfile.new: $!\n";
	unless ($debug) {
		print "Installing $mapfile\n" if $verbose;
		chmod (0644, $nmapfile);
		rename ($nmapfile, $mapfile) 
			or die "Rename $nmapfile to $mapfile failed: $!\n";
	}
}


# crsmap only generated if the file already exists
if (-f $cfile) {
	open (NCRS, ">$cfile.new") or die "Unable to write $cfile.new: $!\n";
	print NCRS join "\n", @cdata;
	print NCRS "\n";
	close (NCRS) or die "unable to close $cfile.new: $!\n";
	chmod (0644, "$cfile.new");

	unless ($debug) {
		print "Installing $cfile\n" if $verbose;
		rename ("$cfile.new", $cfile)
			or die "Rename $cfile.new to $cfile failed: $!\n";
	}
}

open (LU, ">$lastupdatefile.new")
	or die "Unable to write $lastupdatefile.new: $!\n";
print LU $lastchange;
close (LU);
chmod (0644, "$lastupdatefile.new");
unless ($debug) {
	print "Updating $lastupdatefile\n" if $verbose;
	rename ("$lastupdatefile.new", $lastupdatefile);
}
exit 0;

#
# Subroutines
#

sub parseproto {
	my ($protofile) = @_;
	my @protolines;
	open (my $pfh, "$protodir/$protofile") or die "Unable to read $protodir/$protofile: $!\n";
	print "Reading $protofile\n" if $debug;
	while (<$pfh>) {
		chomp;
		next if /^#/;
		if (/^%include (\S+)/) {
			my $incfile = $1;
			print "Including $protodir/$incfile\n" if $debug;
			my $RAinclines = parseproto ($incfile);
			@protolines = (@protolines, @$RAinclines);
			next;
		}
		push (@protolines, $_);
	}
	return \@protolines;
}

sub timeout {
	die "Alarm went off after $alarmdelay seconds\n";
}
